import pytest
import os

from pathlib import Path
from math import isclose
from imlib.general.system import ensure_directory_exists

import cellfinder.tools.system as system
from imlib.general.exceptions import CommandLineInputError

data_dir = Path("tests", "data")
background_im_dir = os.path.join(data_dir, "background")


def test_get_subdirectories():
    subdirs = system.get_subdirectories(data_dir)
    assert len(subdirs) == 10
    assert Path(data_dir / "cells") in subdirs
    assert Path(data_dir / "nii") in subdirs

    subdir_names = system.get_subdirectories(data_dir, names_only=True)
    assert len(subdir_names) == 10
    assert "cells" in subdir_names
    assert "nii" in subdir_names


def test_get_number_of_files_in_dir():
    assert system.get_number_of_files_in_dir(background_im_dir) == 26


def write_file_single_size(directory, file_size):
    with open(os.path.join(directory, str(file_size)), "wb") as fout:
        fout.write(os.urandom(file_size))


def test_check_path_exists(tmpdir):
    num = 10
    tmpdir = str(tmpdir)

    assert system.check_path_exists(os.path.join(tmpdir))
    no_exist_dir = os.path.join(tmpdir, "i_dont_exist")
    with pytest.raises(FileNotFoundError):
        assert system.check_path_exists(no_exist_dir)

    write_file_single_size(tmpdir, num)
    assert system.check_path_exists(os.path.join(tmpdir, str(num)))
    with pytest.raises(FileNotFoundError):
        assert system.check_path_exists(os.path.join(tmpdir, "20"))


def test_catch_input_file_error(tmpdir):
    tmpdir = str(tmpdir)
    # check no error is raised:
    system.catch_input_file_error(tmpdir)

    no_exist_dir = os.path.join(tmpdir, "i_dont_exist")
    with pytest.raises(CommandLineInputError):
        system.catch_input_file_error(no_exist_dir)


def test_ensure_directory_exists():
    home = os.path.expanduser("~")
    exist_dir = os.path.join(home, ".cellfinder_test_dir")
    ensure_directory_exists(exist_dir)
    assert os.path.exists(exist_dir)
    os.rmdir(exist_dir)


def test_memory_in_bytes():
    memory_detection_tolerance = 1  # byte

    assert isclose(
        system.memory_in_bytes(1, "kb"),
        1000,
        abs_tol=memory_detection_tolerance,
    )
    assert isclose(
        system.memory_in_bytes(1.2, "MB"),
        1200000,
        abs_tol=memory_detection_tolerance,
    )
    assert isclose(
        system.memory_in_bytes(0.00065, "gb"),
        650000,
        abs_tol=memory_detection_tolerance,
    )
    assert isclose(
        system.memory_in_bytes(0.000000000234, "TB"),
        234,
        abs_tol=memory_detection_tolerance,
    )
    assert isclose(
        system.memory_in_bytes(1000, "pb"),
        10 ** 18,
        abs_tol=memory_detection_tolerance,
    )

    with pytest.raises(NotImplementedError):
        system.memory_in_bytes(1000, "ab")



from __future__ import absolute_import

cimport cython

from .Visitor cimport CythonTransform, TreeVisitor

cdef class ControlBlock:
     cdef public set children
     cdef public set parents
     cdef public set positions
     cdef public list stats
     cdef public dict gen
     cdef public set bounded

     # Big integer bitsets
     cdef public object i_input
     cdef public object i_output
     cdef public object i_gen
     cdef public object i_kill
     cdef public object i_state

     cpdef bint empty(self)
     cpdef detach(self)
     cpdef add_child(self, block)

cdef class ExitBlock(ControlBlock):
     cpdef bint empty(self)

cdef class NameAssignment:
    cdef public bint is_arg
    cdef public bint is_deletion
    cdef public object lhs
    cdef public object rhs
    cdef public object entry
    cdef public object pos
    cdef public set refs
    cdef public object bit
    cdef public object inferred_type

cdef class AssignmentList:
    cdef public object bit
    cdef public object mask
    cdef public list stats

cdef class AssignmentCollector(TreeVisitor):
    cdef list assignments

@cython.final
cdef class ControlFlow:
     cdef public set blocks
     cdef public set entries
     cdef public list loops
     cdef public list exceptions

     cdef public ControlBlock entry_point
     cdef public ExitBlock exit_point
     cdef public ControlBlock block

     cdef public dict assmts

     cpdef newblock(self, ControlBlock parent=*)
     cpdef nextblock(self, ControlBlock parent=*)
     cpdef bint is_tracked(self, entry)
     cpdef bint is_statically_assigned(self, entry)
     cpdef mark_position(self, node)
     cpdef mark_assignment(self, lhs, rhs, entry)
     cpdef mark_argument(self, lhs, rhs, entry)
     cpdef mark_deletion(self, node, entry)
     cpdef mark_reference(self, node, entry)

     @cython.locals(block=ControlBlock, parent=ControlBlock, unreachable=set)
     cpdef normalize(self)

     @cython.locals(bit=object, assmts=AssignmentList,
                    block=ControlBlock)
     cpdef initialize(self)

     @cython.locals(assmts=AssignmentList, assmt=NameAssignment)
     cpdef set map_one(self, istate, entry)

     @cython.locals(block=ControlBlock, parent=ControlBlock)
     cdef reaching_definitions(self)

cdef class Uninitialized:
     pass

cdef class Unknown:
    pass


cdef class MessageCollection:
    cdef set messages


@cython.locals(dirty=bint, block=ControlBlock, parent=ControlBlock,
               assmt=NameAssignment)
cdef check_definitions(ControlFlow flow, dict compiler_directives)

@cython.final
cdef class ControlFlowAnalysis(CythonTransform):
    cdef object gv_ctx
    cdef object constant_folder
    cdef set reductions
    cdef list env_stack
    cdef list stack
    cdef object env
    cdef ControlFlow flow
    cdef object object_expr
    cdef bint in_inplace_assignment

    cpdef mark_assignment(self, lhs, rhs=*)
    cpdef mark_position(self, node)



"""
prep
==================
Functions to prepare files and directories needed for other functions
"""


import os
import logging
from fancylog import fancylog
from pathlib import Path

from imlib.general.system import ensure_directory_exists, get_num_processes
from imlib.image.metadata import define_pixel_sizes
from imlib.general.exceptions import CommandLineInputError
from imlib.general.config import get_config_obj
from imlib.source import source_files

import amap.download.cli as atlas_download
from amap.cli import check_atlas_install

import cellfinder.tools.tf as tf_tools
import cellfinder as program_for_log
import cellfinder.tools.parser as parser
from cellfinder.download import models as model_download
from cellfinder.download.download import amend_cfg
from cellfinder.download.cli import temp_dir_path
from cellfinder.tools import tools, system


def check_input_arg_existance(args):
    """
    Does a simple check to ensure that input files/paths exist. Prevents a typo
    from causing cellfinder to only run partway. Doesn't check for validity
    etc, just existance.
    :param args: Cellfinder input arguments
    """
    check_list = [args.background_planes_path[0]]
    check_list = check_list + args.signal_planes_paths

    for path in check_list:
        if path is not None:
            system.catch_input_file_error(path)


class Paths:
    """
    A single class to hold all file paths that cellfinder may need. Any paths
    prefixed with "tmp__" refer to internal intermediate steps, and will be
    deleted if "--debug" is not used.
    """

    # TODO: Maybe transfer to a config file.
    def __init__(self, output_dir):
        self.output_dir = output_dir

    def make_reg_paths(self):
        self.registration_output_folder = os.path.join(
            self.output_dir, "registration"
        )
        self.downsampled_background = self.make_reg_path("downsampled.nii")
        self.tmp__downsampled_filtered = self.make_reg_path(
            "downsampled_filtered.nii"
        )
        self.registered_atlas_path = self.make_reg_path("registered_atlas.nii")
        self.hemispheres_atlas_path = self.make_reg_path(
            "registered_hemispheres.nii"
        )
        self.volume_csv_path = self.make_reg_path("volumes.csv")

        self.tmp__affine_registered_atlas_brain_path = self.make_reg_path(
            "affine_registered_atlas_brain.nii"
        )
        self.tmp__freeform_registered_atlas_brain_path = self.make_reg_path(
            "freeform_registered_atlas_brain.nii"
        )
        self.tmp__inverse_freeform_registered_atlas_brain_path = self.make_reg_path(
            "inverse_freeform_registered_brain.nii"
        )

        self.registered_atlas_img_path = self.make_reg_path(
            "registered_atlas.nii"
        )
        self.registered_hemispheres_img_path = self.make_reg_path(
            "registered_hemispheres.nii"
        )

        self.affine_matrix_path = self.make_reg_path("affine_matrix.txt")
        self.invert_affine_matrix_path = self.make_reg_path(
            "invert_affine_matrix.txt"
        )

        self.control_point_file_path = self.make_reg_path(
            "control_point_file.nii"
        )
        self.inverse_control_point_file_path = self.make_reg_path(
            "inverse_control_point_file.nii"
        )

        (
            self.tmp__affine_log_file_path,
            self.tmp__affine_error_path,
        ) = self.compute_reg_log_file_paths("affine")
        (
            self.tmp__freeform_log_file_path,
            self.tmp__freeform_error_file_path,
        ) = self.compute_reg_log_file_paths("freeform")
        (
            self.tmp__inverse_freeform_log_file_path,
            self.tmp__inverse_freeform_error_file_path,
        ) = self.compute_reg_log_file_paths("inverse_freeform")
        (
            self.tmp__segmentation_log_file,
            self.tmp__segmentation_error_file,
        ) = self.compute_reg_log_file_paths("segment")
        (
            self.tmp__invert_affine_log_file,
            self.tmp__invert_affine_error_file,
        ) = self.compute_reg_log_file_paths("invert_affine")

    def make_channel_specific_paths(self):
        self.cells_file_path = os.path.join(self.output_dir, "cells.xml")
        self.tmp__cubes_output_dir = os.path.join(self.output_dir, "cubes")
        self.classification_out_file = os.path.join(
            self.output_dir, "cell_classification.xml"
        )
        self.standard_space_output_folder = os.path.join(
            self.output_dir, "standard_space"
        )
        self.cells_in_standard_space = os.path.join(
            self.standard_space_output_folder, "cells_in_standard_space.xml"
        )
        self.figures_dir = os.path.join(self.output_dir, "figures")

    def make_figures_paths(self):
        self.heatmap = os.path.join(self.figures_dir, "heatmap.nii")

    def make_invert_cell_position_paths(self):
        self.tmp__deformation_field = os.path.join(
            self.standard_space_output_folder, "deformation_field.nii"
        )
        self.tmp__deformation_log_file_path = os.path.join(
            self.standard_space_output_folder, "deformation.log"
        )
        self.tmp__deformation_error_path = os.path.join(
            self.standard_space_output_folder, "deformation.err"
        )

    def make_reg_path(self, basename):
        """
        Compute the absolute path of the destination file to
        self.registration_output_folder.

        :param str basename:
        :return: The path
        :rtype: str
        """
        return os.path.join(self.registration_output_folder, basename)

    def compute_reg_log_file_paths(self, basename):
        """
        Compute the path of the log and err file for the step corresponding
        to basename

        :param str basename:
        :return: log_file_path, error_file_path
        """

        log_file_template = os.path.join(
            self.registration_output_folder, "{}.log"
        )
        error_file_template = os.path.join(
            self.registration_output_folder, "{}.err"
        )
        log_file_path = log_file_template.format(basename)
        error_file_path = error_file_template.format(basename)
        return log_file_path, error_file_path


def prep_cellfinder_general():
    args = parser.cellfinder_parser().parse_args()
    args = define_pixel_sizes(args)
    check_input_arg_existance(args)

    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)

    args.paths = Paths(args.output_dir)
    args.paths.make_reg_paths()

    fancylog.start_logging(
        args.output_dir,
        program_for_log,
        variables=[args, args.paths],
        verbose=args.debug,
        log_header="CELLFINDER LOG",
    )

    what_to_run = CalcWhatToRun(args)
    args.signal_ch_ids, args.background_ch_id = check_and_return_ch_ids(
        args.signal_ch_ids, args.background_ch_id, args.signal_planes_paths
    )
    return args, what_to_run


def check_and_return_ch_ids(signal_ids, background_id, signal_channel_list):
    """
    If channel IDs are provided (via CLI), then they are checked for
    suitability (i.e. they are unique).
    If channel IDs are not given, then unique IDs will be generated

    Assumes only 1 background channel (for registration)

    :param list signal_ids: List of ID ints, or None.
    :param int background_id: ID int, or None.
    :param list signal_channel_list: Input files
    :return: list of signal channel IDs, and background channel ID
    """

    # TODO: make this more general, so it gives informative error messages
    #  when called from cellfinder.extract.cli

    num_signal_channels = len(signal_channel_list)

    if signal_ids is None:
        if background_id is None:
            signal_ids = list(range(num_signal_channels))
            background_id = num_signal_channels  # max(signal_ids) + 1
            pass
        else:
            signal_start = background_id + 1
            signal_end = signal_start + num_signal_channels
            signal_ids = list(range(signal_start, signal_end))
            pass
    else:
        check_correct_number_signal_channels(signal_ids, num_signal_channels)
        check_signal_overlap(signal_ids)
        if background_id is None:
            background_id = max(signal_ids) + 1
            pass
        else:
            # N.B. check_id_overlap needs bg_ch to be iterable (e.g. a list)
            check_sig_bg_id_overlap(signal_ids, [background_id])
    return signal_ids, background_id


def check_correct_number_signal_channels(signal_ids, num_signal_channels):
    num_signal_ids = len(signal_ids)
    if num_signal_channels != num_signal_ids:
        logging.error(
            "Number of signal channel IDs ({}) does not match the number of "
            "signal channels ({})".format(num_signal_ids, num_signal_channels)
        )

        raise CommandLineInputError(
            "Number of signal channel IDs ({}) does not match the number of "
            "signal channels ({})".format(num_signal_ids, num_signal_channels)
        )


def check_signal_overlap(signal_ids):
    signal_unique, repeated_signal_ch = tools.check_unique_list(signal_ids)
    if not signal_unique:
        logging.error(
            "Non-unique values for '--signal-channel-ids' ({}) given. ID:"
            " {} given more than once.".format(signal_ids, repeated_signal_ch)
        )
        raise CommandLineInputError(
            "Non-unique values for '--signal-channel-ids' ({}) given. ID:"
            " {} given more than once.".format(signal_ids, repeated_signal_ch)
        )


def check_sig_bg_id_overlap(signal_ids, background_id):
    common_id_result, intersection = tools.common_member(
        signal_ids, background_id
    )

    if common_id_result:
        logging.error(
            "Non-unique values for '--signal-channel-ids' ({}) and for "
            "'--background-channel-id' ({}) given. ID: {} was given in"
            " both.".format(signal_ids, background_id, intersection)
        )
        raise CommandLineInputError(
            "Non-unique values for '--signal-channel-ids' ({}) and for "
            "'--background-channel-id' ({}) given. ID: {} was given in"
            " both.".format(signal_ids, background_id, intersection)
        )


class CalcWhatToRun:
    """
    Class to (hopefully) simplify what should and shouldn't be run
    """

    def __init__(self, args):
        self.detect = True
        self.classify = True
        self.register = True
        self.summarise = True
        self.figures = True
        self.standard_space = True

        # order is important
        self.cli_options(args)
        self.hierarchy()
        self.existence(args)

    def update(self, args):
        self.cli_options(args)
        self.hierarchy()
        self.existence(args)
        self.channel_specific_update(args)

    def cli_options(self, args):
        self.detect = not args.no_detection
        self.classify = not args.no_classification

        self.register = args.register
        self.summarise = args.summarise
        self.figures = args.figures

        self.standard_space = not args.no_standard_space

    def hierarchy(self):
        if self.summarise or self.figures:
            self.classify = True
            self.register = True

        if not self.detect or not self.classify:
            self.standard_space = False

    def existence(self, args):
        if os.path.exists(
            os.path.join(
                args.paths.registration_output_folder, "registered_atlas.nii"
            )
        ):
            logging.warning(
                "Registered atlas exists, assuming " "already run. Skipping."
            )
            self.register = False
        else:
            # If registration hasn't happened (and it's not going to), then
            # don't carry out standard space analysis
            if not self.register and not self.figures:
                self.standard_space = False

    def channel_specific_update(self, args):
        if os.path.exists(args.paths.cells_file_path):
            logging.warning(
                "Initial detection file exists (cells.xml), "
                "assuming already run. Skipping."
            )
            self.detect = False

        if os.path.exists(args.paths.classification_out_file):
            logging.warning(
                "Cell classification file "
                "(cell_classification.xml)"
                " exists, assuming already run. Skipping cell detection, "
                "and classification"
            )
            self.classify = False
            self.detect = False

        if os.path.exists(
            os.path.join(args.output_dir, "summary_cell_counts.csv")
        ):
            logging.warning(
                "Summary file exists, assuming already run. "
                "skipping everything."
            )
            self.summarise = False
            self.detect = False
            self.classify = False

        if os.path.exists(args.paths.figures_dir) and (
            not os.listdir(args.paths.figures_dir) == []
        ):

            logging.warning(
                "Figures directory not empty, assuming already run. "
                "Skipping detection, and classification."
            )
            self.figures = False
            self.detect = False
            self.classify = False

        if os.path.exists(args.paths.cells_in_standard_space):
            logging.warning(
                "Cells in standard space xml file exists, assuming already "
                "run. Skipping detection and classification."
            )
            self.detect = False
            self.classify = False
            self.standard_space = False


def prep_registration(args, sample_name="amap"):
    logging.info("Checking whether the atlas exists")
    _, atlas_files_exist = check_atlas_install(args.registration_config)
    atlas_dir = args.install_path / "atlas"
    if not atlas_files_exist:
        logging.warning("Atlas does not exist, downloading.")
        atlas_download.atlas_download(
            args.atlas, atlas_dir, args.download_path
        )
        amend_cfg(new_atlas_folder=atlas_dir, atlas=args.atlas)

    args.target_brain_path = args.background_planes_path[0]
    args.sample_name = sample_name
    logging.debug("Making registration directory")
    ensure_directory_exists(args.paths.registration_output_folder)

    additional_images_downsample = {}
    for idx, images in enumerate(args.signal_planes_paths):
        channel = args.signal_ch_ids[idx]
        additional_images_downsample[f"channel_{channel}"] = images

    return args, additional_images_downsample


def check_atlas_install(cfg_file_path=None):
    """
    Checks whether the atlas directory exists, and whether it's empty or not.
    :return: Whether the directory exists, and whether the files also exist
    """
    # TODO: make more sophisticated, check for all files that might be needed
    dir_exists = False
    files_exist = False
    logging.info(cfg_file_path)
    if cfg_file_path is None:
        cfg_file_path = source_files.source_custom_config_cellfinder()
    else:
        pass

    if os.path.exists(cfg_file_path):
        config_obj = get_config_obj(cfg_file_path)
        atlas_conf = config_obj["atlas"]
        atlas_directory = atlas_conf["base_folder"]
        if os.path.exists(atlas_directory):
            dir_exists = True
            if not os.listdir(atlas_directory) == []:
                files_exist = True

    return dir_exists, files_exist


def prep_atlas_conf(args):
    """
    Used by cellfinder_run to source the correct registration configuration
    and also by cellfinder_count_summary to get atlas pixel sizes.

    :param args: Args possibly including "registration_config"
    :return: args: Args with atlas configuration for registration
    """
    if args.atlas_config is None:
        if (
            hasattr(args, "registration_config")
            and args.registration_config is not None
        ):
            args.atlas_config = args.registration_config
        else:
            # can get atlas pixel sizes from registration config
            args.atlas_config = source_files.source_custom_config_cellfinder()
    return args


def prep_classification(args):
    n_processes = get_num_processes(min_free_cpu_cores=args.n_free_cpus)
    prep_tensorflow(n_processes)
    args = prep_models(args)
    return args


def prep_training(args):
    n_processes = get_num_processes(min_free_cpu_cores=args.n_free_cpus)
    prep_tensorflow(n_processes)
    args = prep_models(args)
    return args


def prep_tensorflow(max_threads):
    tf_tools.set_tf_threads(max_threads)
    tf_tools.allow_gpu_memory_growth()


def prep_models(args):
    ## if no model or weights, set default weights
    if args.trained_model is None and args.model_weights is None:
        logging.debug("No model or weights supplied, so using the default")

        config_file = source_files.source_custom_config_cellfinder()
        if not Path(config_file).exists():
            logging.debug("Custom config does not exist, downloading models")
            model_path = model_download.main(args.model, args.install_path)
            amend_cfg(new_model_path=model_path)

        model_weights = get_model_weights(config_file)
        if model_weights is not "" and Path(model_weights).exists():
            args.model_weights = model_weights
        else:
            logging.debug("Model weights do not exist, downloading")
            model_path = model_download.main(args.model, args.install_path)
            amend_cfg(new_model_path=model_path)
            model_weights = get_model_weights(config_file)
            args.model_weights = model_weights
    return args


def get_model_weights(config_file):
    logging.debug(f"Reading config file: {config_file}")
    config_obj = get_config_obj(config_file)
    model_conf = config_obj["model"]
    model_weights = model_conf["model_path"]
    return model_weights


def prep_channel_specific_general(args, what_to_run):
    args.paths.output_dir = args.output_dir
    args.paths.make_channel_specific_paths()
    what_to_run.update(args)
    return args, what_to_run


def prep_candidate_detection(args):
    if args.save_planes:
        args.plane_directory = os.path.join(
            args.output_dir, "processed_planes"
        )
        ensure_directory_exists(args.plane_directory)
    else:
        args.plane_directory = None  # FIXME: remove this fudge

    return args


def standard_space_prep(args):
    ensure_directory_exists(args.paths.standard_space_output_folder)
    args.paths.make_invert_cell_position_paths()
    return args


def figures_prep(args):
    ensure_directory_exists(args.paths.figures_dir)
    args.paths.make_figures_paths()
    return args



import os
import sys

from imlib.IO.cells import get_cells
from cellfinder.utils.xml_scale import main as cellfinder_xml_scale_run

cells_dir = os.path.join("tests", "data", "cells")
orig_xml_path = os.path.join(cells_dir, "cells.xml")
half_scale_scaled_xml_path = os.path.join(
    cells_dir, "cells_rescaled_05_05_1.xml"
)
order_magnitude_scaled_xml_path = os.path.join(
    cells_dir, "cells_rescaled_10_100_1000.xml"
)

SCALED_XML_FILE_NAME = "cells_rescaled.xml"


def run_xml_scale(xml_file, x_scale, y_scale, z_scale, output_dir):
    cellfinder_xml_scale_args = [
        "cellfinder_xml_scale",
        xml_file,
        "-x",
        str(x_scale),
        "-y",
        str(y_scale),
        "-z",
        str(z_scale),
        "-o",
        str(output_dir),
    ]

    sys.argv = cellfinder_xml_scale_args
    cellfinder_xml_scale_run()

    scaled_cells = get_cells(os.path.join(output_dir, SCALED_XML_FILE_NAME))
    return scaled_cells


def test_xml_scale_cli(tmpdir):
    scaled_cells = run_xml_scale(orig_xml_path, 0.5, 0.5, 1, tmpdir)
    assert scaled_cells == get_cells(half_scale_scaled_xml_path)

    scaled_cells = run_xml_scale(orig_xml_path, 10, 100, 1000, tmpdir)
    assert scaled_cells == get_cells(order_magnitude_scaled_xml_path)



import os
from collections import deque
from concurrent.futures import ProcessPoolExecutor
import logging
from datetime import datetime

import numpy as np
from skimage import transform
from numpy.linalg.linalg import LinAlgError
from math import floor
from tifffile import tifffile
from tqdm import tqdm
from imlib.general.system import get_num_processes
from imlib.cells.cells import group_cells_by_z
from imlib.general.numerical import is_even

from cellfinder.tools import image_processing as img_tools
from cellfinder.tools import system


class StackSizeError(Exception):
    pass


class Cube(object):
    class CubeBoundaryError(Exception):
        pass

    def __init__(
        self,
        cell,
        channel,
        stacks,
        x_pix_um=1,
        y_pix_um=1,
        x_pix_um_network=1,
        y_pix_um_network=1,
        final_depth=20,
        width=50,
        height=50,
        depth=20,
        interpolation_order=0,
    ):
        self.cell = cell
        self.channel = channel
        self.stack = stacks[self.channel]
        self.scale_cubes = False
        self.scale_cubes_in_z = False
        self.interpolation_order = interpolation_order
        self.rescaling_factor_x = 1
        self.rescaling_factor_y = 1

        if x_pix_um != x_pix_um_network:
            self.rescaling_factor_x = x_pix_um_network / x_pix_um
            self.scale_cubes = True
        if y_pix_um != y_pix_um_network:
            self.rescaling_factor_y = y_pix_um_network / y_pix_um
            self.scale_cubes = True

        # This should be decided earlier
        if depth != final_depth:
            self.rescaling_factor_z = final_depth / depth
            self.scale_cubes_in_z = True

        self.width = width
        self.height = height
        self.depth = depth
        self.final_depth = depth

        self.empty = False
        self.data = self.extract_cube(self.stack)

    def __str__(self):  # TODO: 0 pad (when changing classification)
        prefix = "pCell"
        return "{}z{}y{}x{}Ch{}.tif".format(
            prefix, self.z, self.y, self.x, self.channel
        )

    @property
    def x(self):
        return self.cell.x

    @property
    def y(self):
        return self.cell.y

    @property
    def z(self):
        return self.cell.z

    @property
    def stack_x_max(self):
        return self.stack.shape[1]

    @property
    def stack_y_max(self):
        return self.stack.shape[0]

    @property
    def stack_z_max(self):
        return self.stack.shape[2]

    def extract_cube(self, stack):
        """

        :param np.array stack:
        :return:
        """
        try:
            self.check_boundaries()
        except Cube.CubeBoundaryError:
            return self.make_blank_data()

        half_width, half_height, x0, x1, y0, y1 = self.get_boundaries()
        data = stack[y0:y1, x0:x1, :]
        if data.ndim < 3:
            logging.error(
                "Error occurred during extraction of cube {}, "
                "original shape: {}, dtype: {}".format(
                    self, data.shape, data.dtype
                )
            )
            return self.make_blank_data()

        out = []
        # TODO: combine these scalings
        for i in range(self.depth):
            if self.scale_cubes:
                try:
                    # To match classification 1um rescaling
                    inv_rescale_x = 1 / self.rescaling_factor_x
                    inv_rescale_y = 1 / self.rescaling_factor_y
                    plane = transform.rescale(
                        data[:, :, i],
                        scale=(inv_rescale_y, inv_rescale_x),
                        order=self.interpolation_order,
                        preserve_range=True,
                        multichannel=False,
                        mode="constant",
                        anti_aliasing=False,
                    )
                except LinAlgError as err:
                    logging.error(
                        f"Error occurred during rescale of cube {self}, "
                        f"original shape: {data[:, :, i].shape}, dtype: "
                        f"{data[:, :, i].dtype}; {err}."
                    )

                    return self.make_blank_data()
            else:
                plane = data[:, :, i]

            plane = self.adjust_size(plane)

            out.append(plane)
        out = np.array(out, dtype=np.uint16)

        if self.scale_cubes_in_z:
            try:
                # To match classification cube scaling in z
                # TODO: presumably this will have rounding errors
                out = transform.rescale(
                    out,
                    scale=(self.rescaling_factor_z, 1, 1),
                    order=self.interpolation_order,
                    preserve_range=True,
                    multichannel=False,
                    mode="constant",
                    anti_aliasing=False,
                )
                out = out.astype(np.uint16)
            except LinAlgError as err:
                logging.error(
                    f"Error occurred during z-rescale of cube {self}, "
                    f"from depth: {self.depth} to final depth  "
                    f"{self.final_depth} ; {err}"
                )
                print.error(
                    f"Error occurred during z-rescale of cube {self}, "
                    f"from depth: {self.depth} to final depth  "
                    f"{self.final_depth} ; {err}"
                )
                return self.make_blank_data()
        return out

    def adjust_size(self, plane):
        plane = self.crop_to_correct_size(plane)
        plane = self.pad_to_correct_size(plane)
        return plane

    def pad_to_correct_size(self, plane):
        return img_tools.pad_center_2d(
            plane, x_size=self.width, y_size=self.height
        )

    def crop_to_correct_size(self, plane):
        return img_tools.crop_center_2d(
            plane, crop_x=self.width, crop_y=self.height
        )

    def get_boundaries(self):
        if self.scale_cubes:
            half_width = int(round(self.width / 2 * self.rescaling_factor_x))
            half_height = int(round(self.height / 2 * self.rescaling_factor_y))
        else:
            half_width = int(round(self.width / 2))
            half_height = int(round(self.height / 2))

        x0 = int(self.x - half_width)
        x1 = int(self.x + half_width)
        y0 = int(self.y - half_height)
        y1 = int(self.y + half_height)
        return half_width, half_height, x0, x1, y0, y1

    def make_blank_data(self):
        self.empty = True
        return np.zeros((self.depth, self.width, self.height), dtype=np.uint16)

    def check_boundaries(self):
        out_bounds_message = (
            "Cell candidate (x:{}, y:{}, z:{}) too close to the edge of the"
            " image for cube extraction."
        ).format(self.x, self.y, self.z)

        if self.stack_z_max < self.depth:
            error_msg = "Stack has {} planes but cube needs {}".format(
                self.stack_z_max, self.depth
            )
            logging.error(error_msg)

            raise Cube.CubeBoundaryError(error_msg)
        boundaries = self.get_boundaries()
        half_height, half_width, x0, x1, y0, y1 = boundaries
        if x0 < 0 or y0 < 0:
            logging.warning(out_bounds_message)
            raise Cube.CubeBoundaryError(out_bounds_message)

        # WARNING: dimensions inverted (numpy y=dim0, x=dim1)
        elif x1 > self.stack_x_max or y1 > self.stack_y_max:
            logging.warning(out_bounds_message)
            raise Cube.CubeBoundaryError(out_bounds_message)


def save_cubes(
    cells,
    planes_paths,
    planes_to_read,
    planes_shape,
    x_pix_um,
    y_pix_um,
    x_pix_um_network,
    y_pix_um_network,
    num_planes_for_cube=20,
    cube_width=50,
    cube_height=50,
    cube_depth=20,
    thread_id=0,
    output_dir="",
    save_empty_cubes=False,
):
    """

    :param cells:
    :param planes_paths:
    :param planes_to_read:
    :param planes_shape:
    :param x_pix_um:
    :param y_pix_um:
    :param x_pix_um_network:
    :param y_pix_um_network:
    :param num_planes_for_cube:
    :param cube_width:
    :param cube_height:
    :param cube_depth:
    :param thread_id:
    :param output_dir:
    :param save_empty_cubes:
    :return:
    """
    channels = list(planes_paths.keys())
    stack_shape = planes_shape + (num_planes_for_cube,)
    stacks = {}
    planes_queues = {}
    for ch in channels:
        stacks[ch] = np.zeros(stack_shape, dtype=np.uint16)
        planes_queues[ch] = deque(maxlen=num_planes_for_cube)
    for plane_idx in tqdm(planes_to_read, desc="Thread: {}".format(thread_id)):
        for ch in channels:
            plane_path = planes_paths[ch][plane_idx]
            planes_queues[ch].append(tifffile.imread(plane_path))
            if len(planes_queues[ch]) == num_planes_for_cube:
                if is_even(num_planes_for_cube):
                    cell_z = int(plane_idx - num_planes_for_cube / 2 + 1)
                else:
                    cell_z = int(
                        plane_idx - floor(num_planes_for_cube) / 2 + 1
                    )

                for j, plane in enumerate(planes_queues[ch]):
                    stacks[ch][:, :, j] = plane

                # ensures no cube_depth planes at the end
                planes_queues[ch].popleft()
                # required since we provide all cells
                # TODO: if len(planes_queues[ch])
                #  < num_planes_for_cube -1: break
                for cell in cells[cell_z]:
                    cube = Cube(
                        cell,
                        ch,
                        stacks,
                        x_pix_um=x_pix_um,
                        y_pix_um=y_pix_um,
                        x_pix_um_network=x_pix_um_network,
                        y_pix_um_network=y_pix_um_network,
                        final_depth=cube_depth,
                        width=cube_width,
                        height=cube_height,
                        depth=num_planes_for_cube,
                    )
                    if not cube.empty or (cube.empty and save_empty_cubes):
                        tifffile.imsave(
                            os.path.join(output_dir, str(cube)), cube.data
                        )


def get_ram_requirement_per_process(
    single_image_path, cube_depth, num_channels=2, copies=2
):
    """
    Calculates how much RAM is needed per CPU core for cube extraction. Used
    later to set the number of CPU cores to be used.

    RAM requirement is currently:
    image_size x n_cores x n_channels x cube_depth x copies

    :param single_image_path: A single 2D image, used to find the size.
    :param num_channels: Number of channels to be extracted
    :param cube_depth: Depth of the cube to be extracted (z-planes)
    to use.
    :param int copies: How many copies of the data are loaded at any time
    :return int: memory requirement for one CPU core
    """

    logging.debug(
        "Determining how much RAM is needed for each CPU for cube extraction"
    )
    file_size = os.path.getsize(single_image_path)
    total_mem_need = file_size * num_channels * cube_depth * copies

    logging.debug(
        "File size: {:.2f} GB, "
        "number of channels: {}, "
        "cube depth: {}, "
        "Therefore theoretically {:.2f} GB is needed per CPU core."
        "".format(
            (file_size / (1024 ** 3)),
            num_channels,
            cube_depth,
            (total_mem_need / (1024 ** 3)),
        )
    )

    return total_mem_need


def main(
    cells,
    cubes_output_dir,
    planes_paths,
    cube_depth,
    cube_width,
    cube_height,
    x_pixel_um,
    y_pixel_um,
    z_pixel_um,
    x_pixel_um_network,
    y_pixel_um_network,
    z_pixel_um_network,
    max_ram,
    n_free_cpus=4,
    save_empty_cubes=False,
):

    start_time = datetime.now()

    if z_pixel_um != z_pixel_um_network:
        plane_scaling_factor = z_pixel_um_network / z_pixel_um
        num_planes_needed_for_cube = round(cube_depth * plane_scaling_factor)
    else:
        num_planes_needed_for_cube = cube_depth

    if num_planes_needed_for_cube > len(planes_paths[0]):
        raise StackSizeError(
            "The number of planes provided is not sufficient "
            "for any cubes to be extracted. Please check the "
            "input data"
        )

    first_plane = tifffile.imread(list(planes_paths.values())[0][0])

    planes_shape = first_plane.shape
    brain_depth = len(list(planes_paths.values())[0])

    # TODO: use to assert all centre planes processed
    center_planes = sorted(list(set([cell.z for cell in cells])))

    # REFACTOR: rename (clashes with different meaning of planes_to_read below)
    planes_to_read = np.zeros(brain_depth, dtype=np.bool)

    if is_even(num_planes_needed_for_cube):
        half_nz = num_planes_needed_for_cube // 2
        # WARNING: not centered because even
        for p in center_planes:
            planes_to_read[p - half_nz : p + half_nz] = 1
    else:
        half_nz = num_planes_needed_for_cube // 2
        # centered
        for p in center_planes:
            planes_to_read[p - half_nz : p + half_nz + 1] = 1

    planes_to_read = np.where(planes_to_read)[0]

    if not planes_to_read.size:
        logging.error(
            f"No planes found, you need at the very least "
            f"{num_planes_needed_for_cube} "
            f"planes to proceed (i.e. cube z size)"
            f"Brain z dimension is {brain_depth}.",
            stack_info=True,
        )
        raise ValueError(
            f"No planes found, you need at the very least "
            f"{num_planes_needed_for_cube} "
            f"planes to proceed (i.e. cube z size)"
            f"Brain z dimension is {brain_depth}."
        )
    # TODO: check if needs to flip args.cube_width and args.cube_height
    cells_groups = group_cells_by_z(cells)

    # copies=2 is set because at all times there is a plane queue (deque)
    # and an array passed to `Cube`
    ram_per_process = get_ram_requirement_per_process(
        planes_paths[0][0], num_planes_needed_for_cube, copies=2,
    )
    n_processes = get_num_processes(
        min_free_cpu_cores=n_free_cpus,
        ram_needed_per_process=ram_per_process,
        n_max_processes=len(planes_to_read),
        fraction_free_ram=0.2,
        max_ram_usage=system.memory_in_bytes(max_ram, "GB"),
    )
    # TODO: don't need to extract cubes from all channels if
    #  n_signal_channels>1
    with ProcessPoolExecutor(max_workers=n_processes) as executor:
        n_planes_per_chunk = len(planes_to_read) // n_processes
        for i in range(n_processes):
            start_idx = i * n_planes_per_chunk
            end_idx = (
                start_idx + n_planes_per_chunk + num_planes_needed_for_cube - 1
            )
            if end_idx > planes_to_read[-1]:
                end_idx = None
            sub_planes_to_read = planes_to_read[start_idx:end_idx]

            executor.submit(
                save_cubes,
                cells_groups,
                planes_paths,
                sub_planes_to_read,
                planes_shape,
                x_pixel_um,
                y_pixel_um,
                x_pixel_um_network,
                y_pixel_um_network,
                num_planes_for_cube=num_planes_needed_for_cube,
                cube_width=cube_width,
                cube_height=cube_height,
                cube_depth=cube_depth,
                thread_id=i,
                output_dir=cubes_output_dir,
                save_empty_cubes=save_empty_cubes,
            )

    total_cubes = system.get_number_of_files_in_dir(cubes_output_dir)
    time_taken = datetime.now() - start_time
    logging.info(
        "All cubes ({}) extracted in: {}".format(total_cubes, time_taken)
    )



import pytest
import random

import numpy as np

import cellfinder.tools.tools as tools


a = [1, "a", 10, 30]
b = [30, 10, "c", "d"]

test_2d_img = np.array([[1, 2, 10, 100], [5, 25, 300, 1000], [1, 0, 0, 125]])
validate_2d_img = np.array(
    [
        [65.535, 131.07, 655.35, 6553.5],
        [327.675, 1638.375, 19660.5, 65535],
        [65.535, 0, 0, 8191.875],
    ]
)


def test_get_max_value():
    num = random.randint(0, 100)
    assert 255 == tools.get_max_value(np.array(num, dtype=np.uint8))
    assert 65535 == tools.get_max_value(np.array(num, dtype=np.uint16))


def test_union():
    ab = [1, "a", 10, 30, "c", "d"]
    assert set(ab) == set(tools.union(a, b))


def test_check_unique_list():
    assert (True, []) == tools.check_unique_list(a)
    repeating_list = [1, 2, 3, 3, "dog", "cat", "dog"]
    assert (False, [3, "dog"]) == tools.check_unique_list(repeating_list)


def test_common_member():
    assert (True, [10, 30]) == tools.common_member(a, b)


def test_get_number_of_bins_nd():
    array_tuple = (100, 1000, 5000)
    assert tools.get_number_of_bins_nd(array_tuple, 12) == (8, 83, 416)

    shape_dict = {"x": 20, "y": 200}
    assert tools.get_number_of_bins_nd(shape_dict, 8) == (2, 25)

    with pytest.raises(NotImplementedError):
        tools.get_number_of_bins_nd([200, 300, 400], 10)


def test_interchange_np_fiji_coordinates():
    tuple_in = (100, 200, 300)
    assert tools.interchange_np_fiji_coordinates(tuple_in) == (200, 100, 300)


def test_swap_elements_list():
    list_in = [1, 2, 3, 4]
    assert tools.swap_elements_list(list_in, 1, 2) == [1, 3, 2, 4]


def test_is_any_list_overlap():
    assert tools.is_any_list_overlap(a, b)
    assert not tools.is_any_list_overlap(a, [2, "b", (1, 2, 3)])



import numpy as np
import logging

from cellfinder.detect.filters.volume_filters.ball_filter import BallFilter
from cellfinder.detect.filters.volume_filters.structure_detection import (
    CellDetector,
    get_structure_centre_wrapper,
)


class StructureSplitException(Exception):
    pass


# def points_to_coords(cell_points):
#     xs = cell_points[:, 0]
#     ys = cell_points[:, 1]
#     zs = cell_points[:, 2]
#     return xs, ys, zs
#
#
# def file_to_coords(file_path):
#     cell_points = np.load(file_path)
#     return points_to_coords(cell_points)


def get_shape(xs, ys, zs):
    # +1 because difference. TEST:
    shape = [int((dim.max() - dim.min()) + 1) for dim in (xs, ys, zs)]
    return shape


def coords_to_volume(xs, ys, zs, ball_radius=1):
    ball_diameter = ball_radius * 2
    # Expanded to ensure the ball fits even at the border
    expanded_shape = [
        dim_size + ball_diameter for dim_size in get_shape(xs, ys, zs)
    ]
    volume = np.zeros(expanded_shape, dtype=np.uint16)

    x_min, y_min, z_min = xs.min(), ys.min(), zs.min()

    relative_xs = np.array((xs - x_min + ball_radius), dtype=np.int64)
    relative_ys = np.array((ys - y_min + ball_radius), dtype=np.int64)
    relative_zs = np.array((zs - z_min + ball_radius), dtype=np.int64)

    # OPTIMISE: vectorize
    for rel_x, rel_y, rel_z in zip(relative_xs, relative_ys, relative_zs):
        volume[rel_x, rel_y, rel_z] = 65534
    return volume


def ball_filter_imgs(
    volume, threshold_value, soma_centre_value, ball_xy_size=3, ball_z_size=3
):
    # OPTIMISE: reuse ball filter instance

    good_tiles_mask = np.ones((1, 1, volume.shape[2]), dtype=np.uint8)

    layer_width, layer_height = volume.shape[:2]

    bf = BallFilter(
        layer_width,
        layer_height,
        ball_xy_size,
        ball_z_size,
        overlap_fraction=0.8,
        tile_step_width=layer_width,
        tile_step_height=layer_height,
        threshold_value=threshold_value,
        soma_centre_value=soma_centre_value,
    )
    cell_detector = CellDetector(
        layer_width, layer_height, start_z=ball_z_size // 2
    )

    # FIXME: hard coded type
    ball_filtered_volume = np.zeros(volume.shape, dtype=np.uint16)
    for z in range(volume.shape[2]):
        bf.append(volume[:, :, z].astype(np.uint16), good_tiles_mask[:, :, z])
        if bf.ready:
            bf.walk()
            middle_plane = bf.get_middle_plane()
            ball_filtered_volume[:, :, z] = middle_plane[:]
            # DEBUG: TEST: transpose
            cell_detector.process(middle_plane.copy())
    return ball_filtered_volume, cell_detector.get_cell_centres()


def iterative_ball_filter(volume, n_iter=10):
    ns = []
    centres = []

    threshold_value = 65534
    soma_centre_value = 65535

    vol = volume.copy()  # TODO: check if required

    for i in range(n_iter):
        vol, cell_centres = ball_filter_imgs(
            vol, threshold_value, soma_centre_value
        )
        vol -= 1
        n_structures = len(cell_centres)
        ns.append(n_structures)
        centres.append(cell_centres)
        if n_structures == 0:
            break
    return ns, centres


def check_centre_in_cuboid(centre, max_coords):
    """
    Checks whether a coordinate is in a cuboid
    :param centre: x,y,z coordinate
    :param max_coords: far corner of cuboid
    :return: True if within cuboid, otherwise False
    """
    relative_coords = np.array([centre[k] for k in ("x", "y", "z")])
    if (relative_coords > max_coords).all():
        logging.info(
            'Relative coordinates "{}" exceed maximum volume '
            'dimension of "{}"'.format(relative_coords, max_coords)
        )
        return False
    else:
        return True


def split_cells(cell_points, relative=False, outlier_keep=False):
    orig_centre = get_structure_centre_wrapper(cell_points)

    xs = np.array([p["x"] for p in cell_points])  # TODO: use dataframe
    ys = np.array([p["y"] for p in cell_points])
    zs = np.array([p["z"] for p in cell_points])

    orig_corner = {
        "x": orig_centre["x"] - (orig_centre["x"] - xs.min()),
        "y": orig_centre["y"] - (orig_centre["y"] - ys.min()),
        "z": orig_centre["z"] - (orig_centre["z"] - zs.min()),
    }
    relative_orig_centre = {
        k: orig_centre[k] - orig_corner[k] for k in ("x", "y", "z")
    }

    original_bounding_cuboid_shape = get_shape(xs, ys, zs)

    ball_radius = 1
    vol = coords_to_volume(xs, ys, zs, ball_radius=ball_radius)

    # centres is a list of lists of centres (1 list of centres per ball run)
    ns, centres = iterative_ball_filter(vol)
    ns.insert(0, 1)
    centres.insert(0, [relative_orig_centre])

    best_iteration = ns.index(max(ns))

    # TODO: put constraint on minimum centres distance ?
    relative_centres = centres[best_iteration]

    if not outlier_keep:
        # TODO: change to checking whether in original cluster shape
        original_max_coords = np.array(original_bounding_cuboid_shape)
        relative_centres = [
            x
            for x in relative_centres
            if check_centre_in_cuboid(x, original_max_coords)
        ]

    if relative:
        return relative_centres
    else:
        absolute_centres = []
        # FIXME: extract functionality
        for relative_centre in relative_centres:
            absolute_centre = {
                "x": orig_corner["x"] + relative_centre["x"],
                "y": orig_corner["y"] + relative_centre["y"],
                "z": orig_corner["z"] + relative_centre["z"],
            }
            absolute_centres.append(absolute_centre)

        return absolute_centres



import logging
import numpy as np
import tensorflow as tf

from random import shuffle
from tifffile import tifffile
from skimage.io import imread
from scipy.ndimage import zoom
from tensorflow.python.keras.utils.data_utils import Sequence
from imlib.IO.cells import get_cells
from imlib.cells.cells import group_cells_by_z
from imlib.general.numerical import is_even

from cellfinder.extract.extract_cubes import StackSizeError
from cellfinder.classify.augment import AugmentationParameters, augment


class CubeGeneratorFromFile(Sequence):
    """
    Reads cubes (defined as e.g. xml, csv) from raw data to pass to
    keras.Model.fit_generator() or keras.Model.predict_generator()

    If augment=True, each augmentation selected has an "augment_likelihood"
    chance of beingapplied to each cube
    """

    # TODO: shuffle within (and maybe between) batches
    # TODO: limit workers based on RAM

    def __init__(
        self,
        cells_file,
        signal_planes,
        background_planes,
        batch_size=16,
        x_pixel_um=1,
        y_pixel_um=1,
        z_pixel_um=5,
        x_pixel_um_network=1,
        y_pixel_um_network=1,
        z_pixel_um_network=5,
        cube_width=50,
        cube_height=50,
        cube_depth=20,
        channels=2,  # No other option currently
        classes=2,
        train=False,
        augment=False,
        augment_likelihood=0.1,
        flip_axis=[0, 1, 2],
        rotate_max_axes=[1, 1, 1],  # degrees
        # scale=[0.5, 2],  # min, max
        translate=[0.05, 0.05, 0.05],
        shuffle=False,
        interpolation_order=2,
    ):
        self.cells_file = cells_file
        self.signal_planes = signal_planes
        self.background_planes = background_planes
        self.batch_size = batch_size
        self.x_pixel_um = x_pixel_um
        self.y_pixel_um = y_pixel_um
        self.z_pixel_um = z_pixel_um
        self.x_pixel_um_network = x_pixel_um_network
        self.y_pixel_um_network = y_pixel_um_network
        self.z_pixel_um_network = z_pixel_um_network
        self.cube_width = cube_width
        self.cube_height = cube_height
        self.cube_depth = cube_depth
        self.channels = channels
        self.classes = classes

        self.train = train
        self.augment = augment
        self.augment_likelihood = augment_likelihood
        self.flip_axis = flip_axis
        self.rotate_max_axes = rotate_max_axes
        # self.scale = scale
        self.translate = translate
        self.shuffle = shuffle
        self.interpolation_order = interpolation_order

        self.scale_cubes = False

        self.rescaling_factor_x = 1
        self.rescaling_factor_y = 1
        self.rescaled_cube_width = self.cube_width
        self.rescaled_cube_height = self.cube_height

        self.__check_image_paths()
        self.__get_image_size()
        self.__load_cells()
        self.__check_z_scaling()
        self.__check_in_plane_scaling()
        self.__remove_outlier_cells()
        self.__get_batches()
        if shuffle:
            self.on_epoch_end()

    def __check_image_paths(self):
        if len(self.signal_planes) != len(self.background_planes):
            raise ValueError(
                f"Number of signal images ({len(self.signal_planes)} does not"
                f"match the number of background images "
                f"({len(self.background_planes)}"
            )

    def __get_image_size(self):
        self.image_z_size = len(self.signal_planes)
        first_plane = tifffile.imread(self.signal_planes[0])
        self.image_height, self.image_width = first_plane.shape

    def __load_cells(self):
        self.cells = get_cells(self.cells_file)
        if not self.cells:
            logging.error(
                f"No cells found, exiting. "
                f"Please check the file: {self.cells_file}"
            )
            raise ValueError(
                f"No cells found, exiting. "
                f"Please check the file: {self.cells_file}"
            )

    def __check_in_plane_scaling(self):
        if self.x_pixel_um != self.x_pixel_um_network:
            self.rescaling_factor_x = self.x_pixel_um_network / self.x_pixel_um
            self.rescaled_cube_width = (
                self.cube_width * self.rescaling_factor_x
            )
            self.scale_cubes = True
        if self.y_pixel_um != self.y_pixel_um_network:
            self.rescaling_factor_y = self.y_pixel_um_network / self.y_pixel_um
            self.rescaled_cube_height = (
                self.cube_height * self.rescaling_factor_y
            )
            self.scale_cubes = True

    def __check_z_scaling(self):
        if self.z_pixel_um != self.z_pixel_um_network:
            plane_scaling_factor = self.z_pixel_um_network / self.z_pixel_um
            self.num_planes_needed_for_cube = round(
                self.cube_depth * plane_scaling_factor
            )
        else:
            self.num_planes_needed_for_cube = self.cube_depth

        if self.num_planes_needed_for_cube > self.image_z_size:
            raise StackSizeError(
                "The number of planes provided is not sufficient "
                "for any cubes to be extracted. Please check the "
                "input data"
            )

    def __remove_outlier_cells(self):
        """
        Remove cells that won't get extracted (i.e too close to the edge)
        """
        self.cells = [cell for cell in self.cells if self.extractable(cell)]

    def extractable(self, cell):
        x0, x1, y0, y1, z0, z1 = self.__get_boundaries()
        if (
            cell.z < z0
            or cell.z > z1
            or cell.x < x0
            or cell.x > x1
            or cell.y < y0
            or cell.y > y1
        ):
            return False
        else:
            return True

    def __get_boundaries(self):
        x0 = int(round((self.cube_width / 2) * self.rescaling_factor_x))
        x1 = int(round(self.image_width - x0))

        y0 = int(round((self.cube_height / 2) * self.rescaling_factor_y))
        y1 = int(round(self.image_height - y0))

        z0 = int(round(self.num_planes_needed_for_cube / 2))
        z1 = self.image_z_size - z0
        return x0, x1, y0, y1, z0, z1

    def __get_batches(self):
        self.cells_groups = group_cells_by_z(self.cells)
        # TODO: add optional shuffling of each group here
        self.batches = []
        for centre_plane in self.cells_groups.keys():
            cells_per_plane = self.cells_groups[centre_plane]
            for i in range(0, len(cells_per_plane), self.batch_size):
                self.batches.append(cells_per_plane[i : i + self.batch_size])

        self.ordered_cells = []
        for batch in self.batches:
            for cell in batch:
                self.ordered_cells.append(cell)

    def __len__(self):
        """
        Number of batches
        :return: Number of batches per epoch
        """
        return len(self.batches)

    def __getitem__(self, index):
        """
        Generates a single batch of cubes
        :param index:
        :return:
        """

        cell_batch = self.batches[index]
        signal_stack, background_stack = self.__get_stacks(index)
        images = self.__generate_cubes(
            cell_batch, signal_stack, background_stack
        )

        if self.train:
            batch_labels = [cell.type - 1 for cell in cell_batch]
            batch_labels = tf.keras.utils.to_categorical(
                batch_labels, num_classes=self.classes
            )
            return images, batch_labels
        else:
            return images

    def __get_stacks(self, index):
        centre_z = self.batches[index][0].z

        half_cube_depth = self.num_planes_needed_for_cube // 2
        min_plane = centre_z - half_cube_depth

        if is_even(self.num_planes_needed_for_cube):
            # WARNING: not centered because even
            max_plane = centre_z + half_cube_depth
        else:
            # centered
            max_plane = centre_z + half_cube_depth + 1

        signal_stack = np.empty(
            (
                self.num_planes_needed_for_cube,
                self.image_height,
                self.image_width,
            )
        )
        background_stack = np.empty_like(signal_stack)
        for plane, plane_path in enumerate(
            self.signal_planes[min_plane:max_plane]
        ):
            signal_stack[plane] = tifffile.imread(plane_path)

        for plane, plane_path in enumerate(
            self.background_planes[min_plane:max_plane]
        ):
            background_stack[plane] = tifffile.imread(plane_path)

        return signal_stack, background_stack

    def __generate_cubes(self, cell_batch, signal_stack, background_stack):
        number_images = len(cell_batch)
        images = np.empty(
            (
                (number_images,)
                + (self.cube_height, self.cube_width, self.cube_depth)
                + (self.channels,)
            )
        )

        for idx, cell in enumerate(cell_batch):
            images = self.__populate_array_with_cubes(
                images, idx, cell, signal_stack, background_stack
            )

        return images

    def __populate_array_with_cubes(
        self, images, idx, cell, signal_stack, background_stack
    ):
        if self.augment:
            self.augmentation_parameters = AugmentationParameters(
                self.flip_axis,
                self.translate,
                self.rotate_max_axes,
                self.interpolation_order,
                self.augment_likelihood,
            )
        images[idx, :, :, :, 0] = self.__get_oriented_image(cell, signal_stack)
        images[idx, :, :, :, 1] = self.__get_oriented_image(
            cell, background_stack
        )
        return images

    def __get_oriented_image(self, cell, image_stack):
        x0 = int(round(cell.x - (self.rescaled_cube_width / 2)))
        x1 = int(x0 + self.rescaled_cube_width)
        y0 = int(round(cell.y - (self.rescaled_cube_height / 2)))
        y1 = int(y0 + self.rescaled_cube_height)
        image = image_stack[:, y0:y1, x0:x1]
        image = np.moveaxis(image, 0, 2)

        if self.augment:
            # scale to isotropic, but don't scale back
            image = augment(
                self.augmentation_parameters, image, scale_back=False
            )

        pixel_scalings = [
            self.cube_height / image.shape[0],
            self.cube_width / image.shape[1],
            self.cube_depth / image.shape[2],
        ]

        # TODO: ensure this is always the correct size
        image = zoom(image, pixel_scalings, order=self.interpolation_order)
        return image

    def on_epoch_end(self):
        """
        Shuffle data for each epoch
        :return: Shuffled indexes
        """
        shuffle(self.batches)


class CubeGeneratorFromDisk(Sequence):
    """
    Reads in cubes from a list of paths for keras.Model.fit_generator() or
    keras.Model.predict_generator()

    If augment=True, each augmentation selected has a 50/50 chance of being
    applied to each cube
    """

    def __init__(
        self,
        signal_list,
        background_list,
        labels=None,  # only if training or validating
        batch_size=16,
        shape=(50, 50, 20),
        channels=2,
        classes=2,
        shuffle=False,
        augment=False,
        augment_likelihood=0.1,
        flip_axis=[0, 1, 2],
        rotate_max_axes=[45, 45, 45],  # degrees
        # scale=[0.5, 2],  # min, max
        translate=[0.2, 0.2, 0.2],
        train=False,  # also return labels
        interpolation_order=2,
    ):
        self.im_shape = shape
        self.batch_size = batch_size
        self.labels = labels
        self.signal_list = signal_list
        self.background_list = background_list
        self.channels = channels
        self.classes = classes
        self.augment = augment
        self.augment_likelihood = augment_likelihood
        self.flip_axis = flip_axis
        self.rotate_max_axes = rotate_max_axes
        # self.scale = scale
        self.translate = translate
        self.train = train
        self.interpolation_order = interpolation_order
        self.indexes = np.arange(len(self.signal_list))
        if shuffle:
            self.on_epoch_end()

    # TODO: implement scale and shear

    def on_epoch_end(self):
        """
        Shuffle data for each epoch
        :return: Shuffled indexes
        """
        self.indexes = np.arange(len(self.signal_list))
        np.random.shuffle(self.indexes)

    def __len__(self):
        """
        Number of batches
        :return: Number of batches per epoch
        """
        return int(np.ceil(len(self.signal_list) / self.batch_size))

    def __getitem__(self, index):
        """
        Generates a single batch of cubes
        :param index:
        :return:
        """
        # Generate indexes of the batch
        start_index = index * self.batch_size
        end_index = start_index + self.batch_size
        indexes = self.indexes[start_index:end_index]

        # Get data corresponding to batch
        list_signal_tmp = [self.signal_list[k] for k in indexes]
        list_background_tmp = [self.background_list[k] for k in indexes]

        images = self.__generate_cubes(list_signal_tmp, list_background_tmp)

        if self.train:
            batch_labels = [self.labels[k] for k in indexes]
            batch_labels = tf.keras.utils.to_categorical(
                batch_labels, num_classes=self.classes
            )
            return images, batch_labels
        else:
            return images

    def __generate_cubes(self, list_signal_tmp, list_background_tmp):
        number_images = len(list_signal_tmp)
        images = np.empty(
            ((number_images,) + self.im_shape + (self.channels,))
        )

        for idx, signal_im in enumerate(list_signal_tmp):
            background_im = list_background_tmp[idx]
            images = self.__populate_array_with_cubes(
                images, idx, signal_im, background_im
            )

        return images.astype(np.float16)

    def __populate_array_with_cubes(
        self, images, idx, signal_im, background_im
    ):
        if self.augment:
            self.augmentation_parameters = AugmentationParameters(
                self.flip_axis,
                self.translate,
                self.rotate_max_axes,
                self.interpolation_order,
                self.augment_likelihood,
            )
        images[idx, :, :, :, 0] = self.__get_oriented_image(signal_im)
        images[idx, :, :, :, 1] = self.__get_oriented_image(background_im)

        return images

    def __get_oriented_image(self, image_path):
        # if paths are pathlib objs, skimage only reads one plane
        image = np.moveaxis(imread(image_path), 0, 2)
        if self.augment:
            image = augment(self.augmentation_parameters, image)
        return image



import random

import numpy as np

from cellfinder.tools import image_processing as img_tools


def test_crop_center_2d():
    x_shape = random.randint(2, 100)
    y_shape = random.randint(2, 100)
    img = np.random.rand(y_shape, x_shape)
    assert (
        img == img_tools.crop_center_2d(img, crop_x=x_shape, crop_y=y_shape)
    ).all()

    new_x_shape = random.randint(1, x_shape)
    new_y_shape = random.randint(1, y_shape)
    pad_img = img_tools.crop_center_2d(
        img, crop_x=new_x_shape, crop_y=new_y_shape
    )
    assert (new_y_shape, new_x_shape) == pad_img.shape


def test_pad_centre_2d():
    x_shape = random.randint(2, 100)
    y_shape = random.randint(2, 100)
    img = np.random.rand(y_shape, x_shape)
    assert (
        img == img_tools.pad_center_2d(img, x_size=x_shape, y_size=y_shape)
    ).all()

    new_x_shape = random.randint(x_shape, x_shape * 10)
    new_y_shape = random.randint(y_shape, y_shape * 10)
    pad_img = img_tools.pad_center_2d(
        img, x_size=new_x_shape, y_size=new_y_shape
    )
    assert (new_y_shape, new_x_shape) == pad_img.shape



from libcpp cimport bool
from libcpp.utility cimport pair
from libc.stddef import ptrdiff_t


cdef extern from "<algorithm>" namespace "std" nogil:
    # Non-modifying sequence operations
    bool all_of[Iter, Pred](Iter first, Iter last, Pred pred) except +
    bool any_of[Iter, Pred](Iter first, Iter last, Pred pred) except +
    bool none_of[Iter, Pred](Iter first, Iter last, Pred pred) except +

    void for_each[Iter, UnaryFunction](Iter first, Iter last, UnaryFunction f) except +  # actually returns f

    ptrdiff_t count[Iter, T](Iter first, Iter last, const T& value) except +
    ptrdiff_t count_if[Iter, Pred](Iter first, Iter last, Pred pred) except +

    pair[Iter1, Iter2] mismatch[Iter1, Iter2](
        Iter1 first1, Iter1 last1, Iter2 first2) except +  # other overloads are tricky

    Iter find[Iter, T](Iter first, Iter last, const T& value) except +
    Iter find_if[Iter, Pred](Iter first, Iter last, Pred pred) except +
    Iter find_if_not[Iter, Pred](Iter first, Iter last, Pred pred) except +

    Iter1 find_end[Iter1, Iter2](Iter1 first1, Iter1 last1, Iter2 first2, Iter2 last2) except +
    Iter1 find_end[Iter1, Iter2, BinaryPred](
        Iter1 first1, Iter1 last1, Iter2 first2, Iter2 last2, BinaryPred pred) except +

    Iter1 find_first_of[Iter1, Iter2](Iter1 first1, Iter1 last1, Iter2 first2, Iter2 last2) except +
    Iter1 find_first_of[Iter1, Iter2, BinaryPred](
        Iter1 first1, Iter1 last1, Iter2 first2, Iter2 last2, BinaryPred pred) except +

    Iter adjacent_find[Iter](Iter first, Iter last) except +
    Iter adjacent_find[Iter, BinaryPred](Iter first, Iter last, BinaryPred pred) except +

    Iter1 search[Iter1, Iter2](Iter1 first1, Iter1 last1, Iter2 first2, Iter2 last2) except +
    Iter1 search[Iter1, Iter2, BinaryPred](
        Iter1 first1, Iter1 last1, Iter2 first2, Iter2 last2, BinaryPred pred) except +
    Iter search_n[Iter, Size, T](Iter first1, Iter last1, Size count, const T& value) except +
    Iter search_n[Iter, Size, T, BinaryPred](
        Iter first1, Iter last1, Size count, const T& value, BinaryPred pred) except +

    # Modifying sequence operations
    OutputIt copy[InputIt, OutputIt](InputIt first, InputIt last, OutputIt d_first) except +
    OutputIt copy_if[InputIt, OutputIt, Pred](InputIt first, InputIt last, OutputIt d_first, Pred pred) except +
    OutputIt copy_n[InputIt, Size, OutputIt](InputIt first, Size count, OutputIt result) except +
    Iter2 copy_backward[Iter1, Iter2](Iter1 first, Iter1 last, Iter2 d_last) except +

    OutputIt move[InputIt, OutputIt](InputIt first, InputIt last, OutputIt d_first) except +
    Iter2 move_backward[Iter1, Iter2](Iter1 first, Iter1 last, Iter2 d_last) except +

    void fill[Iter, T](Iter first, Iter last, const T& value) except +
    Iter fill_n[Iter, Size, T](Iter first, Size count, const T& value) except +

    OutputIt transform[InputIt, OutputIt, UnaryOp](
        InputIt first1, InputIt last1, OutputIt d_first, UnaryOp unary_op) except +
    OutputIt transform[InputIt1, InputIt2, OutputIt, BinaryOp](
        InputIt1 first1, InputIt1 last1, InputIt2 first2, OutputIt d_first, BinaryOp binary_op) except +

    void generate[Iter, Generator](Iter first, Iter last, Generator g) except +
    void generate_n[Iter, Size, Generator](Iter first, Size count, Generator g) except +

    Iter remove[Iter, T](Iter first, Iter last, const T& value) except +
    Iter remove_if[Iter, UnaryPred](Iter first, Iter last, UnaryPred pred) except +
    OutputIt remove_copy[InputIt, OutputIt, T](InputIt first, InputIt last, OutputIt d_first, const T& value) except +
    OutputIt remove_copy_if[InputIt, OutputIt, UnaryPred](
        InputIt first, InputIt last, OutputIt d_first, UnaryPred pred) except +

    void replace[Iter, T](Iter first, Iter last, const T& old_value, const T& new_value) except +
    void replace_if[Iter, UnaryPred, T](Iter first, Iter last, UnaryPred pred, const T& new_value) except +
    OutputIt replace_copy[InputIt, OutputIt, T](
        InputIt first, InputIt last, OutputIt d_first, const T& old_value, const T& new_value) except +
    OutputIt replace_copy_if[InputIt, OutputIt, UnaryPred, T](
        InputIt first, InputIt last, OutputIt d_first, UnaryPred pred, const T& new_value) except +

    void swap[T](T& a, T& b) except +  # array overload also works
    Iter2 swap_ranges[Iter1, Iter2](Iter1 first1, Iter1 last1, Iter2 first2) except +
    void iter_swap[Iter](Iter a, Iter b) except +

    void reverse[Iter](Iter first, Iter last) except +
    OutputIt reverse_copy[InputIt, OutputIt](InputIt first, InputIt last, OutputIt d_first) except +

    Iter rotate[Iter](Iter first, Iter n_first, Iter last) except +
    OutputIt rotate_copy[InputIt, OutputIt](InputIt first, InputIt n_first, InputIt last, OutputIt d_first) except +

    Iter unique[Iter](Iter first, Iter last) except +
    Iter unique[Iter, BinaryPred](Iter first, Iter last, BinaryPred p) except +
    OutputIt unique_copy[InputIt, OutputIt](InputIt first, InputIt last, OutputIt d_first) except +
    OutputIt unique_copy[InputIt, OutputIt, BinaryPred](
        InputIt first, InputIt last, OutputIt d_first, BinaryPred pred) except +

    # Partitioning operations
    bool is_partitioned[Iter, Pred](Iter first, Iter last, Pred p) except +
    Iter partition[Iter, Pred](Iter first, Iter last, Pred p) except +
    pair[OutputIt1, OutputIt2] partition_copy[InputIt, OutputIt1, OutputIt2, Pred](
        InputIt first, InputIt last, OutputIt1 d_first_true, OutputIt2 d_first_false, Pred p) except +
    Iter stable_partition[Iter, Pred](Iter first, Iter last, Pred p) except +
    Iter partition_point[Iter, Pred](Iter first, Iter last, Pred p) except +

    # Sorting operations
    bool is_sorted[Iter](Iter first, Iter last) except +
    bool is_sorted[Iter, Compare](Iter first, Iter last, Compare comp) except +

    Iter is_sorted_until[Iter](Iter first, Iter last) except +
    Iter is_sorted_until[Iter, Compare](Iter first, Iter last, Compare comp) except +

    void sort[Iter](Iter first, Iter last) except +
    void sort[Iter, Compare](Iter first, Iter last, Compare comp) except +

    void partial_sort[Iter](Iter first, Iter middle, Iter last) except +
    void partial_sort[Iter, Compare](Iter first, Iter middle, Iter last, Compare comp) except +

    OutputIt partial_sort_copy[InputIt, OutputIt](
        InputIt first, InputIt last, OutputIt d_first, OutputIt d_last) except +
    OutputIt partial_sort_copy[InputIt, OutputIt, Compare](
        InputIt first, InputIt last, OutputIt d_first, OutputIt d_last, Compare comp) except +

    void stable_sort[Iter](Iter first, Iter last) except +
    void stable_sort[Iter, Compare](Iter first, Iter last, Compare comp) except +

    void nth_element[Iter](Iter first, Iter nth, Iter last) except +
    void nth_element[Iter, Compare](Iter first, Iter nth, Iter last, Compare comp) except +

    # Binary search operations (on sorted ranges)
    Iter lower_bound[Iter, T](Iter first, Iter last, const T& value) except +
    Iter lower_bound[Iter, T, Compare](Iter first, Iter last, const T& value, Compare comp) except +

    Iter upper_bound[Iter, T](Iter first, Iter last, const T& value) except +
    Iter upper_bound[Iter, T, Compare](Iter first, Iter last, const T& value, Compare comp) except +

    bool binary_search[Iter, T](Iter first, Iter last, const T& value) except +
    bool binary_search[Iter, T, Compare](Iter first, Iter last, const T& value, Compare comp) except +

    # Other operations on sorted ranges

    # Set operations (on sorted ranges)

    # Heap operations
    void make_heap[Iter](Iter first, Iter last) except +
    void make_heap[Iter, Compare](Iter first, Iter last, Compare comp) except +

    void push_heap[Iter](Iter first, Iter last) except +
    void push_heap[Iter, Compare](Iter first, Iter last, Compare comp) except +

    void pop_heap[Iter](Iter first, Iter last) except +
    void pop_heap[Iter, Compare](Iter first, Iter last, Compare comp) except +

    void sort_heap[Iter](Iter first, Iter last) except +
    void sort_heap[Iter, Compare](Iter first, Iter last, Compare comp) except +

    # Minimum/maximum operations
    Iter min_element[Iter](Iter first, Iter last) except +

    # Comparison operations

    # Permutation operations



# cython: language_level=3

cdef class Action:
    cdef perform(self, token_stream, text)
    cpdef same_as(self, other)

cdef class Return(Action):
    cdef object value
    cdef perform(self, token_stream, text)
    cpdef same_as(self, other)

cdef class Call(Action):
    cdef object function
    cdef perform(self, token_stream, text)
    cpdef same_as(self, other)

cdef class Method(Action):
    cdef str name
    cdef dict kwargs

cdef class Begin(Action):
    cdef object state_name
    cdef perform(self, token_stream, text)
    cpdef same_as(self, other)

cdef class Ignore(Action):
    cdef perform(self, token_stream, text)

cdef class Text(Action):
    cdef perform(self, token_stream, text)





cdef extern from "pythread.h":

    ctypedef void *PyThread_type_lock
    ctypedef void *PyThread_type_sema

    void PyThread_init_thread()
    long PyThread_start_new_thread(void (*)(void *), void *)  # FIXME: legacy
    #unsigned long PyThread_start_new_thread(void (*)(void *), void *)  # returned 'long' before Py3.7
    void PyThread_exit_thread()
    long PyThread_get_thread_ident()  # FIXME: legacy
    #unsigned long PyThread_get_thread_ident()  # returned 'long' before Py3.7

    PyThread_type_lock PyThread_allocate_lock()
    void PyThread_free_lock(PyThread_type_lock)
    int PyThread_acquire_lock(PyThread_type_lock, int mode) nogil
    void PyThread_release_lock(PyThread_type_lock) nogil

    enum:
        # 'mode' in PyThread_acquire_lock()
        WAIT_LOCK    #   1
        NOWAIT_LOCK  #   0

    ctypedef enum PyLockStatus:
        # return values of PyThread_acquire_lock() in CPython 3.2+
        PY_LOCK_FAILURE = 0
        PY_LOCK_ACQUIRED = 1
        PY_LOCK_INTR

    size_t PyThread_get_stacksize()
    int PyThread_set_stacksize(size_t)

    # Thread Local Storage (TLS) API deprecated in CPython 3.7+
    int PyThread_create_key()
    void PyThread_delete_key(int)
    int PyThread_set_key_value(int, void *)
    void * PyThread_get_key_value(int)
    void PyThread_delete_key_value(int key)

    # Cleanup after a fork
    void PyThread_ReInitTLS()

    # Thread Specific Storage (TSS) API in CPython 3.7+ (also backported)
    #ctypedef struct Py_tss_t: pass   # Cython built-in type
    Py_tss_t Py_tss_NEEDS_INIT        # Not normally useful: Cython auto-initialises declared "Py_tss_t" variables.
    Py_tss_t * PyThread_tss_alloc()
    void PyThread_tss_free(Py_tss_t *key)
    int PyThread_tss_is_created(Py_tss_t *key)
    int PyThread_tss_create(Py_tss_t *key)
    void PyThread_tss_delete(Py_tss_t *key)
    int PyThread_tss_set(Py_tss_t *key, void *value)
    void * PyThread_tss_get(Py_tss_t *key)
